unit Organizer;

{ TOrganizer components: Encapsulate the Organizer97 API.
  Version 0.2  Mar/19/1997  (C) 1997 Christoph R. Kirchner
}
{ Users of this unit must accept this disclaimer of  warranty:
    "This unit is supplied as is. The author disclaims all warranties,
    expressed or implied, including, without limitation, the warranties
    of merchantability and of fitness for any purpose.
    The author assumes no liability for damages, direct or
    consequential, which may result from the use of this unit."

  This Unit is donated to the public as public domain.

  This Unit can be freely used and distributed in commercial and
  private environments provided this notice is not modified in any way.

  If you do find this Unit handy and you feel guilty for using such a
  great product without paying someone - sorry.

  Please forward any comments or suggestions to Christoph Kirchner at:
  ckirchner@geocities.com

  Maybe you can find an update of this component at my
  "Delphi Component Collection":
  http://www.geocities.com/SiliconValley/Heights/7874/delphi.htm
}


interface

uses
  Windows, SysUtils, Classes, Forms, OrgApi;

type

  TOSessionOption =
    (OSO_GATEWAYSESSION, OSO_GRAYDIALOGS, OSO_VALIDATE, OSO_NOUI);
  TOSessionOptions = set of TOSessionOption;

  TOSession = class(TComponent)
  private
    FhOrgAPI: HORG_Session;
    FContainerPath: string;
    FUserName: string;
    FPassword: string;
    FOptions: TOSessionOptions;
    FStreamedActive: Boolean;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
  protected
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    property Handle: HORG_Session read FhOrgAPI;
    function APIVersion: string;
  published
    property Active: Boolean read GetActive write SetActive default False;
    property ContainerPath: string
      read FContainerPath write FContainerPath;
    property Options: TOSessionOptions
      read FOptions write FOptions default [];
    property Password: string read FPassword write FPassword;
    property UserName: string read FUserName write FUserName;
  end;

  TOFileOption =
    (foBURSTLOAD, foPRELOADTABLES, foNOPASSWORDBOX);
  TOFileOptions = set of TOFileOption;
  TOAccessLevel =
    (aclNone, aclReader, aclAssistant, aclTrustee, aclOwner);
  TOAccessLevels = set of TOAccessLevel;

  TOFile = class(TComponent)
  private
    FhOrgFile: HORG_File;
    FOSession: TOSession;
    FAccessLevel: TOAccessLevel;   { actual level }
    FAccessLevels: TOAccessLevels; { allowed levels }
    FFileName: string;
    FPassword: string;
    FOptions: TOFileOptions;
    FStreamedActive: Boolean;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    property Handle: HORG_File read FhOrgFile;
    procedure GetSectionList(List: TStrings);
  published
    property Active: Boolean read GetActive write SetActive default False;
    property AccessLevels: TOAccessLevels
      read FAccessLevels write FAccessLevels
      default [aclReader, aclAssistant, aclTrustee, aclOwner];
    property FileName: string read FFileName write FFileName;
    property Options: TOFileOptions
      read FOptions write FOptions default [];
    property OSession: TOSession
      read FOSession write FOSession;
    property Password: string read FPassword write FPassword;
  end;

  TORecord = class;
  TOField = class;

  TOSection = class(TComponent)
  private
    FhOrgSection: HORG_Section;
    FOFile: TOFile;
    FSectionName: string;
    FStreamedActive: Boolean;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    function SectionType: ORGAPISecType;
    function FieldCount: Integer;
    function RecordCount: Integer;
    procedure GetFieldList(List: TStrings);
    function GetFieldName(FieldNum: Integer): string;
    property Handle: HORG_Section read FhOrgSection;
  published
    property Active: Boolean read GetActive write SetActive default False;
    property OFile: TOFile
      read FOFile write FOFile;
    property SectionName: string read FSectionName write FSectionName;
  end;

  TOFieldType =
    (oftLIST, oftBYTE, oftWORD, oftDWORD, oftVIRTUALFILE, oftBOOL,
     oftSTRING, oftSTRINGLIST, oftDATETIME, oftBITMAP, oftMETAFILE,
     oftOLE, oftRTF, oftCOLORREF, oftUNIQID, oftVIRFILE2);

  TORecordState = (orInactive, orBrowse, orEdit, orInsert);

  TORecord = class(TComponent)
  private
    FhOrgRecord: HORG_Record;
    FRecordPosition: Longint;
    FOSection: TOSection;
    FStreamedActive: Boolean;
    FFields: TList;
    FBOF: Boolean;
    FEOF: Boolean;
    FState: TORecordState;
    function GetBOF: Boolean;
    function GetEOF: Boolean;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
    function GetField(Index: Integer): TOField;
    function CreateField(Index: Integer): TOField;
    procedure SetOSection(Value: TOSection);
    procedure InsertAddress;
    procedure RecordIsEmpty;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure CreateFields;
    procedure DestroyFields;
    procedure RemoveField(OField: TOField);
    procedure CheckInactive;
    procedure CheckBrowseMode;
    procedure RecordChanged; virtual;
    function OrgRecordPosition: Longint; { RecordPosition got from OrgAPI }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    function GetFieldValue(FieldNumber: Integer): Pointer;
    function GetFieldType(FieldNumber: Integer): TOFieldType;
    function FieldCount: Integer;
    function RecordCount: Integer;
    function GetFieldName(FieldNum: Integer): string;
    procedure First;
    procedure Last;
    procedure Next;
    procedure Prior;
    procedure Insert;
    procedure Edit;
    procedure Post;
    procedure Delete;
    procedure Cancel;
    property Handle: HORG_Record read FhOrgRecord;
    property BOF: Boolean read GetBOF;
    property EOF: Boolean read GetEOF;
    property RecNo: Longint read FRecordPosition;
    property Fields[Index: Integer]: TOField read GetField;
    property State: TORecordState read FState;
  published
    property Active: Boolean read GetActive write SetActive default False;
    property OSection: TOSection
      read FOSection write SetOSection;
  end;

  TOField = class(TComponent)
  private
    FORecord: TORecord;
    FFieldNumber: Integer;
    FFieldName: string;
    FValidValue: Boolean;
    procedure GetText(var Text: string; DisplayText: Boolean);
    function GetDisplayText: string;
    function GetEditText: string;
  protected
    function GetParentComponent: TComponent; override;
    function HasParent: Boolean; override;
    procedure SetParentComponent(AParent: TComponent); override;
    function GetFieldValue: Pointer;
    procedure AccessError(const TypeName: string);
    procedure CheckInactive;
    function GetAsBoolean: Boolean; virtual;
    function GetAsDateTime: TDateTime; virtual;
    function GetAsInteger: Longint; virtual;
    function GetAsString: string; virtual;
    function GetAsVariant: Variant; virtual;
    procedure SetAsBoolean(Value: Boolean); virtual;
    procedure SetAsDateTime(Value: TDateTime); virtual;
    procedure SetAsInteger(Value: Longint); virtual;
    procedure SetAsString(const Value: string); virtual;
    procedure SetAsVariant(const Value: Variant); virtual;
    procedure SetText(const Value: string); virtual;
    procedure SetVarValue(const Value: Variant); virtual;
    procedure Edit;
  public
    destructor Destroy; override;
    procedure RecordChanged; virtual;
    procedure Clear; virtual;
    property ORecord: TORecord
      read FORecord write FORecord;
    property FieldNumber: Integer read FFieldNumber;
    property Text: string read GetEditText write SetText;
    property FieldName: string read FFieldName;
  end;

  TOStringField = class(TOField)
  private
    FBuffer: string;
  protected
    function GetAsString: string; override;
    procedure SetAsString(const Value: string); override;
  public
    procedure Clear; override;
  end;

  TOStringListField = class(TOStringField)
  protected
    function GetAsString: string; override;
    procedure SetAsString(const Value: string); override;
  public
    procedure Clear; override;
  end;



procedure register;

implementation

function MapFieldType(OrganizerFieldType: ORGFieldType): TOFieldType;
begin
  if (OrganizerFieldType = ORGAPI_STRINGLIST) then
    result := oftSTRINGLIST
  else
    result := TOFieldType(OrganizerFieldType AND $F);
end;


{ TOSession -------------------------------------------------------- }

constructor TOSession.Create(AOwner: TComponent);
begin
  inherited;
  FhOrgAPI := OA_FAILURE;
  FContainerPath := '';
  FUserName := ' ';
  FPassword := '';
  FOptions := [];
  FStreamedActive := false;
end;

destructor TOSession.Destroy;
begin
  Close;
  inherited;
end;

procedure TOSession.Open;
const
  SessionOptions: array[TOSessionOption] of Integer =
   (OAIF_GATEWAYSESSION, OAIF_GRAYDIALOGS, OAIF_VALIDATE, OAIF_NOUI);
var
  WindowHandle: HWND;
  Err: Integer;
  ApiInitFlags: Integer;
  SessionOption: TOSessionOption;
begin
  if (Owner is TForm) then
    WindowHandle := TForm(Owner).Handle
  else
    WindowHandle := 0;
  ApiInitFlags := 0;
  for SessionOption := OSO_GATEWAYSESSION to OSO_NOUI do
    ApiInitFlags := ApiInitFlags + SessionOptions[SessionOption];
  FhOrgAPI := OrgAPIInit(WindowHandle, 0, 0, 0, PChar(FContainerPath),
    PChar(FUserName), PChar(FPassword), ApiInitFlags, Err);
  if (FhOrgAPI = OA_FAILURE) then
    raise EOrganizerError.Create('Unable to open Organizer session');
end;

procedure TOSession.Close;
begin
  if (FhOrgAPI <> OA_FAILURE) then
  begin
    OrgAPITerm(FhOrgAPI);
    FhOrgAPI := OA_FAILURE;
  end;
end;

function TOSession.GetActive: Boolean;
begin
  Result := (FhOrgAPI <> OA_FAILURE);
end;

procedure TOSession.SetActive(Value: Boolean);
begin
  if (csReading in ComponentState) then
  begin
    if Value then FStreamedActive := Value;
  end
  else
    if Active <> Value then
    begin
      if Value then
      begin
        Open;
      end else
      begin
        Close;
      end;
    end;
end;

procedure TOSession.Loaded;
begin
  inherited Loaded;
  try
    if FStreamedActive then Open;
  except
{
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
}
      raise;
  end;
end;

function TOSession.APIVersion: string;
var
  s: array[0..127] of Char;
  p: pchar;
begin
{ SetLength(Result, 127);
  OrgAPIVersion(PChar(Result), 127); }
  s[0] := #0;
  OrgAPIVersion(s, 127);
  Result := s;
end;


{ TOFile ------------------------------------------------------------ }

constructor TOFile.Create(AOwner: TComponent);
begin
  inherited;
  FhOrgFile := OA_FAILURE;
  FFileName := '';
  FPassword := '';
  FOptions := [];
  FStreamedActive := false;
  FOSession := nil;
  FAccessLevel := aclNone;
  FAccessLevels := [aclReader, aclAssistant, aclTrustee, aclOwner];
end;

destructor TOFile.Destroy;
begin
  Close;
  inherited;
end;

procedure TOFile.Open;
const
  OAccessLevels: array [TOAccessLevel] of Integer =
   (0, OACL_READER, OACL_ASSISTANT, OACL_TRUSTEE, OACL_OWNER);
var
  WindowHandle: HWND;
  OpenFlag: Integer;
  TryAccessLevel: TOAccessLevel;
begin
  if (FhOrgFile <> OA_FAILURE) then
    Close;
  if not Assigned(FOSession) then
    raise EOrganizerError.Create('OrganizerSession not set');
  if (Owner is TForm) then
    WindowHandle := TForm(Owner).Handle
  else
    WindowHandle := 0;
  OpenFlag := 0;
  if not (foBURSTLOAD in FOptions) then
    OpenFlag := OpenFlag + OA_NOBURSTLOAD;
  if not (foPRELOADTABLES in FOptions) then
    OpenFlag := OpenFlag + OA_NOPRELOADTABLES;
  if (foNOPASSWORDBOX in FOptions) then
    OpenFlag := OpenFlag + OA_NOPASSWORDBOX;
  for TryAccessLevel := aclOwner downto aclReader do
  begin
    if (TryAccessLevel in FAccessLevels) then
    begin
      FhOrgFile := OrgAPIFileOpen(FOSession.Handle, WindowHandle,
        PChar(FFileName), PChar(FPassword),
        OAccessLevels[TryAccessLevel], OpenFlag);
      if (FhOrgFile <> OA_FAILURE) then
        break;
    end;
  end;
  if (FhOrgFile = OA_FAILURE) then
  begin
    FAccessLevel := aclNone;
    raise EOrganizerError.Create('Unable to open Organizer File');
  end
  else
    FAccessLevel := TryAccessLevel;
end;

procedure TOFile.Close;
begin
  if (FhOrgFile <> OA_FAILURE) then
  begin
    OrgAPIFileClose(FhOrgFile);
    FhOrgFile := OA_FAILURE;
    FAccessLevel := aclNone;
  end;
end;

function TOFile.GetActive: Boolean;
begin
  Result := (FhOrgFile <> OA_FAILURE);
end;

procedure TOFile.SetActive(Value: Boolean);
begin
  if (csReading in ComponentState) then
  begin
    if Value then FStreamedActive := Value;
  end
  else
    if Active <> Value then
    begin
      if Value then
      begin
        Open;
      end else
      begin
        Close;
      end;
    end;
end;

procedure TOFile.Loaded;
begin
  inherited Loaded;
  try
    if FStreamedActive then Open;
  except
{
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
}
      raise;
  end;
end;

procedure TOFile.Notification(
  AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if(Operation = opRemove) and (FOSession <> nil) and
   (AComponent = FOSession) then
  begin
    if Active then
      Close;
    FOSession := nil;
  end;
end;

procedure TOFile.GetSectionList(List: TStrings);
var
  TableNum: Integer;
  TableName: array[0..255] of Char;
begin
  List.Clear;
  TableNum := 1;
  while (OrgAPIEnumerateSectionsAlt(FhOrgFile, TableNum,
                                    PChar(@TableName), 255) = OA_SUCCESS) do
  begin
    List.Add(string(TableName));
  end;
end;



{ TOSection --------------------------------------------------- }

constructor TOSection.Create(AOwner: TComponent);
begin
  inherited;
  FhOrgSection := OA_FAILURE;
  FSectionName := '';
  FStreamedActive := false;
  FOFile := nil;
end;

destructor TOSection.Destroy;
begin
  Close;
  inherited;
end;

procedure TOSection.Open;
begin
  if (FhOrgSection <> OA_FAILURE) then
    Close;
  if not Assigned(FOFile) then
    raise EOrganizerError.Create('OFile not set');
  if (FSectionName = '') then
    raise EOrganizerError.Create('SectionName not set');
  FhOrgSection := OrgAPISectionOpen(FOFile.Handle,
    PChar(FSectionName));
  if (FhOrgSection = OA_FAILURE) then
    raise EOrganizerError.Create('Unable to open Organizer Section');
end;

procedure TOSection.Close;
begin
  if (FhOrgSection <> OA_FAILURE) then
  begin
    OrgAPISectionClose(FhOrgSection);
    FhOrgSection := OA_FAILURE;
  end;
end;

function TOSection.GetActive: Boolean;
begin
  Result := (FhOrgSection <> OA_FAILURE);
end;

procedure TOSection.SetActive(Value: Boolean);
begin
  if (csReading in ComponentState) then
  begin
    if Value then FStreamedActive := Value;
  end
  else
    if Active <> Value then
    begin
      if Value then
      begin
        Open;
      end else
      begin
        Close;
      end;
    end;
end;

procedure TOSection.Loaded;
begin
  inherited Loaded;
  try
    if FStreamedActive then Open;
  except
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

procedure TOSection.Notification(
  AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if(Operation = opRemove) and (FOFile <> nil) and
   (AComponent = FOFile) then
  begin
    if Active then
      Close;
    FOFile := nil;
  end;
end;

function TOSection.FieldCount: Integer;
begin
  if (not Active) then
    result := 0
  else
    result := OrgAPIFieldGetFieldCount(FhOrgSection);
end;

function TOSection.RecordCount: Integer;
begin
  if (not Active) then
    result := 0
  else
    result := OrgAPISectionRecordCount(FhOrgSection);
end;

function TOSection.SectionType: ORGAPISecType;
begin
  if (not Active) then
    result := 0
  else
    result := OrgAPISectionType(FhOrgSection);
end;

procedure TOSection.GetFieldList(List: TStrings);
var
  FieldNum: Integer;
  FieldName: array[0..255] of Char;
begin
  List.Clear;
  FieldNum := 0;
  while (OrgAPIFieldEnumAlt(FhOrgSection, FieldNum,
                            PChar(@FieldName), 255) = OA_SUCCESS) do
  begin
    List.Add(string(FieldName));
  end;
end;

function TOSection.GetFieldName(FieldNum: Integer): string;
var
  FieldName: array[0..255] of Char;
begin
  if (OrgAPIFieldEnumAlt(FhOrgSection, FieldNum,
                         PChar(@FieldName), 255) = OA_SUCCESS) then
    result := string(FieldName)
  else
    result := '';
end;


{ TORecord ---------------------------------------------------------- }

constructor TORecord.Create(AOwner: TComponent);
begin
  inherited;
  FhOrgRecord := OA_FAILURE;
  FStreamedActive := false;
  FOSection := nil;
  FBOF := true;
  FEOF := true;
  FRecordPosition := 1;
  FFields := TList.Create;
  FState := orInactive;
end;

destructor TORecord.Destroy;
begin
  Close;
  DestroyFields;
  FFields.Free;
  inherited;
end;

procedure TORecord.Open;
var
  Position: Integer;
begin
  if (FhOrgRecord <> OA_FAILURE) then
    Close;
  if not Assigned(FOSection) then
    raise EOrganizerError.Create('Organizer Section not set');
  if (RecordCount > 0) then
  begin
    Position := FRecordPosition;
    if (Position > 0) and
       (FOSection.SectionType = ORGAPI_ADDRESS_TYPE) then
      Inc(Position, 28); { Address-section begins with record-number 29 }
    FhOrgRecord := OrgAPIRecordOpen(FOSection.Handle, Position);
    if (FhOrgRecord = OA_FAILURE) then
      raise EOrganizerError.Create('Unable to open Organizer Record');
    FEOF := false;
  end;
  FState := orBrowse;
  if (FFields.Count = 0) then
    CreateFields
  else
    RecordChanged;
end;

procedure TORecord.RecordIsEmpty;
begin
  FhOrgRecord := OA_FAILURE;
  FBOF := true;
  FEOF := true;
  RecordChanged;
end;

procedure TORecord.Close;
begin
  if (FhOrgRecord <> OA_FAILURE) then
  begin
    OrgAPIRecordClose(FhOrgRecord);
    RecordIsEmpty;
  end;
end;

function TORecord.GetActive: Boolean;
begin
  result := (FState <> orInactive);
end;

procedure TORecord.SetActive(Value: Boolean);
begin
  if (csReading in ComponentState) then
  begin
    if Value then FStreamedActive := Value;
  end
  else
    if Active <> Value then
    begin
      if Value then
      begin
        Open;
      end else
      begin
        Close;
      end;
    end;
end;

function TORecord.GetFieldValue(FieldNumber: Integer): Pointer;
begin
  CheckInactive;
  if (FhOrgRecord = OA_FAILURE) then
    result := nil
  else
    result := OrgAPIFieldGet(FhOrgRecord, FieldNumber);
end;

function TORecord.GetFieldType(FieldNumber: Integer): TOFieldType;
var
  OrganizerFieldType: ORGFieldType;
begin
  if not Assigned(FOSection) then
    raise EOrganizerError.Create('Organizer Section not set');
  if not FOSection.Active then
    raise EOrganizerError.Create('Organizer Section not active');
  OrganizerFieldType := OrgAPIFieldGetFieldType(FOSection.Handle, FieldNumber);
  result := MapFieldType(OrganizerFieldType);
end;

procedure TORecord.Notification(
  AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if(Operation = opRemove) and (FOSection <> nil) and
   (AComponent = FOSection) then
  begin
    if Active then
      Close;
    FOSection := nil;
  end;
end;

procedure TORecord.Loaded;
begin
  inherited Loaded;
  FRecordPosition := 1;
  try
    if FStreamedActive then Open;
  except
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

function TORecord.FieldCount: Integer;
begin
  if not Assigned(FOSection) then
    result := 0
  else
    result := FOSection.FieldCount;
end;

function TORecord.RecordCount: Integer;
begin
  if not Assigned(FOSection) then
    result := 0
  else
    result := FOSection.RecordCount;
end;

function TORecord.OrgRecordPosition: Longint;
begin
  if (FhOrgRecord = OA_FAILURE) then
    result := 0
  else
    result := OrgAPIRecordPosition(FhOrgRecord);
end;

function TORecord.GetBOF: Boolean;
begin
  Result := FBOF;
end;

function TORecord.GetEOF: Boolean;
begin
  Result := FEOF;
end;

procedure TORecord.CheckInactive;
begin
  if not Active then
    raise EOrganizerError.Create('Organizer record not active');
end;

procedure TORecord.CheckBrowseMode;
begin
  CheckInactive;
  if (FState = orEdit) or (FState = orInsert) then
    Post;
end;

procedure TORecord.First;
begin
  CheckBrowseMode;
  if (RecordCount > 0) then
  begin
    if (FRecordPosition = 1) then
      FBOF := true
    else
    begin
      Close;
      FRecordPosition := 1;
      Open;
    end;
    FEOF := false;
  end
  else
  begin
    RecordIsEmpty;
    FRecordPosition := 0;
  end;
end;

procedure TORecord.Last;
begin
  CheckBrowseMode;
  if (RecordCount > 0) then
  begin
    if (FRecordPosition <> RecordCount) then
    begin
      Close;
      FRecordPosition := RecordCount;
      Open;
    end;
    FBOF := false;
    FEOF := true;
  end
  else
  begin
    RecordIsEmpty;
    FRecordPosition := 0;
  end;
end;

procedure TORecord.Next;
begin
  CheckBrowseMode;
  if (RecordCount > 0) then
  begin
    if (not EOF) and (FRecordPosition < RecordCount) then
    begin
      Close;
      Inc(FRecordPosition);
      Open;
    end
    else
    begin
      FBOF := false;
      FEOF := true;
    end;
  end
  else
  begin
    RecordIsEmpty;
    FRecordPosition := 0;
  end;
end;

procedure TORecord.Prior;
begin
  CheckBrowseMode;
  if (RecordCount > 0) then
  begin
    if (not BOF) and (FRecordPosition > 1) then
    begin
      Close;
      Dec(FRecordPosition);
      Open;
    end
    else
    begin
      FBOF := true;
      FEOF := false;
    end;
  end
  else
  begin
    RecordIsEmpty;
    FRecordPosition := 0;
  end;
end;

procedure TORecord.RemoveField(OField: TOField);
begin
  OField.ORecord := nil;
  FFields.Remove(OField);
end;

function TORecord.GetField(Index: Integer): TOField;
begin
  Result := FFields[Index];
end;

function TORecord.CreateField(Index: Integer): TOField;
type
  TFieldClass = class of TOField;
const
  FieldClass: array[TOFieldType] of TFieldClass =
   (TOField, TOField, TOField, TOField, TOField, TOField,
    TOStringField, TOStringListField, TOField, TOField, TOField, TOField,
    TOField, TOField, TOField, TOField);
{   (oftLIST, oftBYTE, oftWORD, oftDWORD, oftVIRTUALFILE, oftBOOL,
     oftSTRING, oftSTRINGLIST, oftDATETIME, oftBITMAP, oftMETAFILE, oftOLE,
     oftRTF, oftCOLORREF, oftUNIQID, oftVIRFILE2) }
begin
  Result := FieldClass[GetFieldType(Index)].Create(self);
  try
    Result.ORecord := self;
    Result.FFieldNumber := Index;
    Result.FFieldName := GetFieldName(Index);
  except
    Result.Free;
    raise;
  end;
end;

procedure TORecord.CreateFields;
var
  i: Integer;
begin
  for i := 0 to FieldCount - 1 do
    FFields.Add(CreateField(i));
end;

procedure TORecord.DestroyFields;
var
  Field: TOField;
begin
  while FFields.Count > 0 do
  begin
    Field := FFields.Last;
    RemoveField(Field);
    Field.Free;
  end;
end;

procedure TORecord.SetOSection(Value: TOSection);
var
  WasActive: Boolean;
begin
  if (Value <> FOSection) then
  begin
    WasActive := Active;
    Close;
    DestroyFields;
    FOSection := Value;
    if WasActive and Assigned(FOSection) then
      Open; { calls CreateFields too }
  end;
end;

function TORecord.GetFieldName(FieldNum: Integer): string;
begin
  if Assigned(FOSection) then
    result := FOSection.GetFieldName(FieldNum)
  else
    result := '';
end;

procedure TORecord.RecordChanged;
var
  i: Integer;
begin
  for i := 0 to FFields.Count - 1 do
    TOField(FFields[i]).RecordChanged;
end;

procedure TORecord.Edit;
begin
  CheckInactive;
  if (FState = orBrowse) then
  begin
    FState := orEdit;
    raise EOrganizerError.Create('Edit not implemented yet');
  end;
end;

procedure TORecord.Insert;
var
  i: Integer;
begin
  if (FOSection.SectionType <> ORGAPI_ADDRESS_TYPE) then
    raise EOrganizerError.Create('Insert not implemented yet');
  FState := orInsert;
  for i := 0 to FFields.Count - 1 do
    TOField(FFields[i]).Clear;
end;

procedure TORecord.Post;
begin
  if (FState = orInsert) then
  begin
    case FOSection.SectionType of
      ORGAPI_ADDRESS_TYPE: InsertAddress;
    end;
    FRecordPosition := OrgRecordPosition;
    if (FRecordPosition > 28) and
       (FOSection.SectionType = ORGAPI_ADDRESS_TYPE) then
      Dec(FRecordPosition, 28); { Address-section begins with record-number 29 }
    FBOF := (FRecordPosition = 1);
    FEOF := (FRecordPosition = RecordCount);
    RecordChanged;
  end;
  if (FState = orEdit) then
  begin
    CheckInactive;
    RecordChanged;
  end;
  FState := orBrowse;
end;

procedure TORecord.Cancel;
begin
  if (FState = orInsert) or (FState = orEdit) then
  begin
    FState := orBrowse;
    RecordChanged;
  end;
end;


procedure TORecord.Delete;
begin
  CheckBrowseMode;
  if (OrgAPIRecordDelete(FhOrgRecord) = OA_FAILURE) then
    raise EOrganizerError.Create('Deletion of record failed');
  RecordIsEmpty;
  if (FRecordPosition > RecordCount) then
    FRecordPosition := RecordCount;
  if (FRecordPosition > 0) then
    Open;
end;

procedure TORecord.InsertAddress;
var
  RecAddr: TORGADDRESS;
begin
  FillChar(RecAddr, SizeOf(RecAddr), 0);
  with RecAddr do
  begin
    Title := PChar(TOStringField(Fields[ORGADDR_TITLE]).FBuffer);
    FName := PChar(TOStringField(Fields[ORGADDR_FNAME]).FBuffer);
    LName := PChar(TOStringField(Fields[ORGADDR_LNAME]).FBuffer);
    Position := PChar(TOStringField(Fields[ORGADDR_POSITION]).FBuffer);
    Company := PChar(TOStringField(Fields[ORGADDR_COMPANY]).FBuffer);
    Address := PChar(TOStringField(Fields[ORGADDR_ADDRESS]).FBuffer);
    City := PChar(TOStringField(Fields[ORGADDR_CITY]).FBuffer);
    State := PChar(TOStringField(Fields[ORGADDR_STATE]).FBuffer);
    Zip := PChar(TOStringField(Fields[ORGADDR_ZIP]).FBuffer);
    Country := PChar(TOStringField(Fields[ORGADDR_COUNTRY]).FBuffer);
    Phone := PChar(TOStringField(Fields[ORGADDR_PHONE]).FBuffer);
    Ext := PChar(TOStringField(Fields[ORGADDR_EXT]).FBuffer);
    Fax := PChar(TOStringField(Fields[ORGADDR_FAX]).FBuffer);
    Email := PChar(TOStringField(Fields[ORGADDR_EMAIL]).FBuffer);
    Unused1 := PChar(TOStringField(Fields[ORGADDR_UNUSED1]).FBuffer);
    Unused2 := PChar(TOStringField(Fields[ORGADDR_UNUSED2]).FBuffer);
    Notes := PChar(TOStringField(Fields[ORGADDR_NOTES]).FBuffer);
    HomeSpouse := PChar(TOStringField(Fields[ORGADDR_HOME_SPOUSE]).FBuffer);
    HomeChildren := PChar(TOStringField(Fields[ORGADDR_HOME_CHILDREN]).FBuffer);
    HomeAddress := PChar(TOStringField(Fields[ORGADDR_HOME_ADDRESS]).FBuffer);
    HomeCity := PChar(TOStringField(Fields[ORGADDR_HOME_CITY]).FBuffer);
    HomeState := PChar(TOStringField(Fields[ORGADDR_HOME_STATE]).FBuffer);
    HomeZip := PChar(TOStringField(Fields[ORGADDR_HOME_ZIP]).FBuffer);
    HomeCountry := PChar(TOStringField(Fields[ORGADDR_HOME_COUNTRY]).FBuffer);
    HomePhone := PChar(TOStringField(Fields[ORGADDR_HOME_PHONE]).FBuffer);
    HomeExt := PChar(TOStringField(Fields[ORGADDR_HOME_EXT]).FBuffer);
    HomeFax := PChar(TOStringField(Fields[ORGADDR_HOME_FAX]).FBuffer);
    HomeEmail := PChar(TOStringField(Fields[ORGADDR_HOME_EMAIL]).FBuffer);
    HomeUnused1 := PChar(TOStringField(Fields[ORGADDR_HOME_UNUSED1]).FBuffer);
    HomeUnused2 := PChar(TOStringField(Fields[ORGADDR_HOME_UNUSED2]).FBuffer);
    HomeNotes := PChar(TOStringField(Fields[ORGADDR_HOME_NOTES]).FBuffer);
    lpfConfidential := nil;  { PBoolean  ORGADDR_CONFIDENTIAL }
    Category := PChar(TOStringListField(Fields[ORGADDR_CATEGORY]).FBuffer);
  end;
  OrgAPIRecordClose(FhOrgRecord);
  FhOrgRecord := OrgAPIAddressInsert(
    FOSection.FOFile.Handle, FOSection.Handle, RecAddr);
end;


{ TOField ----------------------------------------------------------- }

destructor TOField.Destroy;
begin
  if FORecord <> nil then
  begin
    FORecord.Close;
    FORecord.RemoveField(Self);
  end;
  inherited Destroy;
end;

function TOField.GetParentComponent: TComponent;
begin
  Result := ORecord;
end;

procedure TOField.SetParentComponent(AParent: TComponent);
begin
  if not (csLoading in ComponentState) then
    ORecord := AParent as TORecord;
end;

function TOField.HasParent: Boolean;
begin
  Result := True;
end;

function TOField.GetFieldValue: Pointer;
begin
  if Assigned(FORecord) and FORecord.Active and
     (FORecord.Handle <> OA_FAILURE) then
    result := OrgAPIFieldGet(FORecord.Handle, FieldNumber)
  else
    result := nil;
end;

procedure TOField.AccessError(const TypeName: string);
begin
  raise EOrganizerError.Create('Wrong field-type: ' + TypeName);
end;

procedure TOField.CheckInactive;
begin
  if Assigned(FORecord) then
    FORecord.CheckInactive;
end;

function TOField.GetAsBoolean: Boolean;
begin
  AccessError('Boolean');
end;

function TOField.GetAsDateTime: TDateTime;
begin
  AccessError('DateTime');
end;

function TOField.GetAsInteger: Longint;
begin
  AccessError('Integer');
end;

function TOField.GetAsString: string;
begin
  AccessError('String');
end;

function TOField.GetAsVariant: Variant;
begin
  AccessError('Variant');
end;

procedure TOField.Clear;
begin
end;

procedure TOField.SetAsBoolean(Value: Boolean);
begin
  AccessError('Boolean');
end;

procedure TOField.SetAsDateTime(Value: TDateTime);
begin
  AccessError('DateTime');
end;

procedure TOField.SetAsInteger(Value: Longint);
begin
  AccessError('Integer');
end;

procedure TOField.SetAsString(const Value: string);
begin
  AccessError('String');
end;

procedure TOField.SetAsVariant(const Value: Variant);
begin
  if TVarData(Value).VType = varNull then
    Clear
  else
    try
      SetVarValue(Value);
    except
      AccessError('Variant');
    end;
end;

procedure TOField.SetVarValue(const Value: Variant);
begin
  AccessError('Variant');
end;

procedure TOField.SetText(const Value: string);
begin
  SetAsString(Value);
end;

procedure TOField.GetText(var Text: string; DisplayText: Boolean);
begin
  Text := GetAsString;
end;

function TOField.GetDisplayText: string;
begin
  Result := '';
  GetText(Result, True);
end;

function TOField.GetEditText: string;
begin
  Result := '';
  GetText(Result, False);
end;

procedure TOField.RecordChanged;
begin
  FValidValue := false;
end;

procedure TOField.Edit;
begin
  if Assigned(FORecord) then
    FORecord.Edit;
  FValidValue := true;
end;


{ TOStringField ------------------------------------------------------------- }

function TOStringField.GetAsString: string;
begin
  if not FValidValue then
  begin
    FBuffer := PChar(GetFieldValue);
    FValidValue := true;
  end;
  result := FBuffer;
end;

procedure TOStringField.SetAsString(const Value: string);
begin
  Edit;
  FBuffer := Value;
end;

procedure TOStringField.Clear;
begin
  Edit;
  Text := '';
end;


{ TOStringListField --------------------------------------------------------- }

const
  FDelimiter = ',';

function TOStringListField.GetAsString: string;
type
  tc = array [0..1023] of Char;
  pc = ^tc;
var
  Value: pc;
  i:Integer;
  Size: Integer;
begin
  if not FValidValue then
  begin
    Size := 0;
    Value := pc(GetFieldValue);
    if Assigned(Value) then
    begin
      while (Value^[Size] <> #0) do
      begin
        inc(Size);
        if (Value^[Size] = #0) then
          inc(Size);
      end;
      Dec(Size, 1);
    end;
    if (Size > 0) then
    begin
      SetLength(FBuffer, Size + 1);
      Move(Value^, FBuffer[1], Size);
      FBuffer[Size + 1] := #0;
    end
    else
      FBuffer := '';
    FValidValue := true;
  end;
  if (FBuffer <> '') then
  begin
    SetLength(result, Length(FBuffer) - 1);
    Move(FBuffer[1], result[1], Length(FBuffer) - 1);
    for i := 1 to Length(FBuffer) - 1 do
      if (result[i] = #0) then
        result[i] := FDelimiter;
  end
  else
    result := '';
end;

procedure TOStringListField.SetAsString(const Value: string);
var
  i:Integer;
begin
  Edit;
  if (Value <> '') then
  begin
    SetLength(FBuffer, Length(Value) + 1);
    Move(Value[1], FBuffer[1], Length(Value));
    for i := 1 to Length(Value) do
      if (FBuffer[i] = FDelimiter) then
        FBuffer[i] := #0;
    FBuffer[Length(Value) + 1] := #0;
  end
  else
    FBuffer := '';
end;

procedure TOStringListField.Clear;
begin
  Edit;
  Text := '';
end;


{ Register ------------------------------------------------------------------ }

procedure Register;
begin
  RegisterComponents('Organizer',
    [TOSession, TOFile, TOSection, TORecord]);
end;


end.
